/****************************************************************************
* FENPRINT-MAIL Vn 1.0  July 2004
* By Mike Eggleston (Fenland Software Ltd)
*
* Copyright (c) 2004 Fenland Software Ltd 
*****************************************************************************/
#define PROGNAME	"FENPRINT-MAIL"
#define DESC		"Fenland Software's Virtual Printing Utility"
#define VERSION		"1.0"
#define COPYRIGHT	"Copyright (c) 2004 Fenland Software Ltd"

#define _BOUNDARY	"--fenprint-mail-boundary_marker"

#include "fenprint.h"
#include <resolv.h>

char *fmts[] = { "AUTO", "PBM", "FAX", "TEXT", "TIFF", "DFAX", "PCX", 
"RAW", "DCX", "JPG", "GIF", "PDF", "PS"};

char *sfxs[] = { "aut", "pbm", "fax", "txt", "tif", "dfx", "pcx", 
"raw", "dcx", "jpg", "gif", "pdf", "ps"};

char *cnt_types[] = {"application", "application", "image",
                     "text", "image", "application",
                     "image", "application", "image",
                     "image", "image", "application",
                     "application"};

char *cnt_subtypes[] = {"octet-stream", "octet-stream", "octet-stream",
                        "plain", "tiff", "octet-stream",
                        "octet-stream", "octet-stream", "octet-stream",
                        "jpeg", "gif", "pdf", "postscript"};


enum iformats { I_AUTO=0, I_PBM=1, I_FAX=2, I_TEXT=3, I_TIFF=4,
		I_DFAX=5, I_PCX=6, I_RAW=7, I_DCX=8, I_JPG=9,
                I_GIF=10, I_PDF=11, I_PS=12} ;

#define IFORMATS { "AUTO", "PBM", "FAX", "TEXT", "TIFF", "DFAX", \
		   "PCX", "RAW", "DCX", "JPG", "GIF", "PDF", "PS" } ;

char sendmail_path[_PATH_LEN + 1] = "";
char reply_to[_PATH_LEN + 1] = "";
char mail_intro[_PATH_LEN + 1] = "";
char mail_title[_PATH_LEN + 1] = "";


void usage(void)
{
  fprintf(stderr, "%s Vn %s\n%s\n", PROGNAME,VERSION,COPYRIGHT);
  fprintf(stderr, "%s\n", DESC);
  fprintf(stderr, "Please see the manuals for operating instructions\n\n");
  exit(0);
}

/* Validate an email address domain */
int valid_email(char *addr)
{
  char *s;
  unsigned char answer[_PATH_LEN + 1];

  s = strchr(addr, '@');
  if(s != NULL)
  {
    s++;
    res_init();
    if(res_query(s,C_ANY,T_ANY,answer,_PATH_LEN) != -1)
      return 1;
  }

  return 0;
}


/****************************************************************************
* Load the configuration
****************************************************************************/
int freadln(FILE *fp,char *str)
{
  if(fgets(str,_PATH_LEN,fp) != NULL)
    if(str[0] != '#')
      return 1;
  return 0;
}


int get_param(char *d_str,char *s_str,char *key,int max)
{
  char tok[_PATH_LEN + 1],param[_PATH_LEN + 1];

  tok[0] = 0;
  param[0] = 0;
  if(sscanf(s_str,"%s %s",tok,param) == 0)
    return 0;

  if(!strcmp(tok,key))
  {
    char *s;

    s = strstr(s_str + strlen(tok),param);
    if(s != NULL) strncpy(d_str,s,max);
    d_str[max] = 0;
    s = strchr(d_str,'\n');
    if(s != NULL) *s = 0;
    s = strchr(d_str,'\r');
    if(s != NULL) *s = 0;

    return 1;
  }

  return 0;
}


int load_conf(void)
{
  FILE *fp;
  char str[_PATH_LEN + 1];
  char substr[_PATH_LEN + 1];

  if((fp = fopen(_CONFIG_FILE, "r")) == NULL)
  {
    fprintf(stderr,"ERROR - Can't open configuration file - %s\n", _CONFIG_FILE);
    return(0);
  }

  while(!feof(fp))
    if(freadln(fp, str))
    {
      if(get_param(substr,str,"sendmail_path:",_PATH_LEN))
        strcpy(sendmail_path,substr);

      if(get_param(substr,str,"reply_to:",_PATH_LEN))
      {
        strcpy(reply_to,"Reply-To: ");
        strcat(reply_to, substr);
      }

      if(get_param(substr,str,"mail_intro:",_PATH_LEN))
        strcpy(mail_intro, substr);

      if(get_param(substr,str,"mail_title:",_PATH_LEN))
        strcpy(mail_title, substr);

    }

  fclose(fp);
  return(1);
}

/****************************************************************/
/* Deduce the file type by scanning buffer p of n bytes. */
/****************************************************************/
int getformat ( uchar *p, int n )
{
  int format = 0 ;

  /* figure out file type if not already set */

  if ( ! format && n < 2 ) {
    format = I_TEXT ;
  } 

  if ( ! format && ! p[0] && ! ( p[1] & 0xe0 ) ) {
    format = I_FAX ;
  } 

  if ( ! format && ! strncmp ( p, "P4", 2 ) ) {
    format = I_PBM ;
  }

  if ( ! format && n >= 128 && p[0] == 0x0a && 
       strchr ("\02\03\05", p[1] ) && p[2] <= 1 ) {
    if ( p[65] != 1 ) {
    } else {
      format = p[2] ? I_PCX : I_PCX ;
    }
  }

/*  if ( ! format && ! strncmp ( p, "%!PS-", 5 ) ) { */
  if ( ! format && ! strncmp ( p, "%!", 2 ) ) {
    format = I_PS ;
  }

  if ( ! format && ( p[0] == 'M' || p[1] == 'I' ) && ( p[1] == p[0] ) ) {
    format = I_TIFF ;
  }
  
  if ( ! format &&
       ( p[0] == 0x3a && p[1] == 0xde && 
	 p[2] == 0x68 && p[3] == 0xb1) ) {
    format = I_DCX ;
  }
  
  if ( ! format &&
       ( p[6] == 0x4a && p[7] == 0x46 && 
	 p[8] == 0x49 && p[9] == 0x46) ) {
    format = I_JPG ;
  }

  if ( ! format && ! strncmp(p, "GIF89", 5) ) {
    format = I_GIF ;
  }

  if ( ! format && ! strncmp(p, "%PDF-", 5) ) {
    format = I_PDF ;
  }

  if ( ! format &&  n ) { /* "90% printable" heuristic */
    int i, nprint = 0 ;
    for ( i=0 ; i<n ; i++ ) {
      if ( isspace ( p[i] ) || isprint ( p[i] ) ) {
	nprint++ ;
      }
    }
    if ( ( nprint / (float) n ) > 0.90 ) {
      format = I_TEXT ;
    }
  }

  return format ;
}

int mail_header(FILE *fp_mail, char *send_to, char *subject)
{
FILE *fp;
char str[_BUFFER_LEN + 1];

  fprintf(fp_mail, "To: %s\n%s\nSubject: %s\nMIME-Version: 1.0\n\
Content-Type: multipart/mixed; boundary=%s\n\n--%s\n\
Content-Type: text/plain; charset=\"US-ASCII\"\n\
Content-Transfer-Encoding: 7bit\n\n\n", send_to, reply_to, subject, 
_BOUNDARY, _BOUNDARY);

  if(fp=fopen(mail_intro, "r"))
  {
    while(fgets(str, _BUFFER_LEN, fp))
      fprintf(fp_mail, "%s",str);

    fclose(fp);
  }

return 0;
}

int doc_header(FILE *fp_mail, char *desc, int format)
{
char fmt[5];
char sfx[4];
char typ[15];
char styp[15];

    strcpy(fmt, fmts[format]);
    strcpy(sfx, sfxs[format]);
    strcpy(typ, cnt_types[format]);
    strcpy(styp, cnt_subtypes[format]);

  fprintf(fp_mail, "--%s\nContent-Type: %s/%s; name=document.%s\n\
Content-Description: %s\n\
Content-Disposition: attachment; filename=document.%s\n\
Content-Transfer-Encoding: BASE64\n\n",
_BOUNDARY, typ, styp, sfx, desc, sfx);

return 1;
}


/* Copy stdin to fp_out while applying base64 (RFC 1521)
   encoding.  This encoding must be applied after the file is
   complete since some output formats (e.g. TIFF) require seeking
   backwards within the (binary) file.  */

int base64encode (FILE *fp_out, int buff[], int length)
{
  int err=0, c, cnt=0 ;
  uchar n=0, m=0, bits=0 ;

  static uchar chartab[65] = 
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ" 
    "abcdefghijklmnopqrstuvwxyz"
    "0123456789+/" ;

  if(cnt < length)
    c = buff[cnt++];
  else
    c = fgetc(stdin);

  while ( c >= 0 ) {
    switch ( n ) {
    case 0:
      putc ( chartab [ c >> 2 ], fp_out ) ;
      bits = c & 0x3 ;
      n = 1 ;
      break ;
    case 1:
      putc ( chartab [ (bits << 4) | ( c >> 4 ) ], fp_out ) ;
      bits = c & 0xf ;
      n = 2 ;
      break ;
    case 2:
      putc ( chartab [ (bits << 2) | ( c >> 6 ) ], fp_out ) ;
      putc ( chartab [ c & 0x3f ], fp_out ) ;
      n = 0 ;
      if ( ++m >= 18 ) {
	putc ( '\n', fp_out ) ;
	m = 0 ;
      }
      break ;
    }

  if(cnt < length)
    c = buff[cnt++];
  else
    c = fgetc(stdin);

  }

  switch ( n ) {
  case 0:
    break ;
  case 1:
    putc ( chartab [ (bits << 4) | ( 0 >> 4 ) ], fp_out ) ;
    putc ( '=', fp_out ) ;
    putc ( '=', fp_out ) ;
    break ;
  case 2 :
    putc ( chartab [ (bits << 2) | ( 0 >> 6 ) ], fp_out ) ;
    putc ( '=', fp_out ) ;
    break ;
  }

  putc ( '\n', fp_out ) ;

  return err ;
}

int send_by_mail(char *send_to, char *subject, char *file_desc)
{
FILE *fp_mail;
int n, format, err;
int buffer[_BUFFER_LEN];
uchar str[_BUFFER_LEN + 1];

  for(n=0; n < _BUFFER_LEN; n++)
  {
    buffer[n] = fgetc(stdin);
    str[n] = (uchar) buffer[n];
  }

  if((format = getformat(str,n)) >= 0)
  {
    sprintf(str,"%s %s",sendmail_path, send_to);
    if(fp_mail = popen(str,"w"))
    {
      mail_header(fp_mail, send_to, subject);
      doc_header(fp_mail, file_desc, format);
      err = base64encode(fp_mail, buffer, n);
      fprintf(fp_mail, "--%s--\n\n",_BOUNDARY);
      pclose(fp_mail);
    }

  return 1;
  }

return 0;
}

int main( int argc, char **argv)
{
char send_to[_PATH_LEN + 1] = " ";
char title[_PATH_LEN + 1] = " ";
char description[_PATH_LEN + 1] = "document";

  if((argc < 2) || (argc > 4))
    usage();

  if(!load_conf())
    return -1;

  strcpy(send_to, argv[1]);

  if(argc > 2)
    strcpy(title, argv[2]);
  else
    strcpy(title, mail_title);

  if(argc > 3)
    strcpy(description, argv[3]);

  if(valid_email(send_to))
    send_by_mail(send_to, title, description);
  else
  {
    fprintf(stderr, "ERROR - Attempt to print to invalid email address\n");
    fprintf(stderr, "ERROR -   %s\n", send_to);
    return -1;
  }
return 0;
}


/*    End of source file for FENPRINT-MAIL    */

